#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging
import sys


def get_custom_logger(logger_name, level, stream=sys.stdout):
    r"""
    Create a custom logger with a standard formatting

    Inputs:
    - logger_name: Name of the logger. You can get the class name as self.__class__.__name__
    - level: logging level (e.g. logging.INFO, logging.DEBUG, etc.)
    - stream: One of sys.stdout or sys.stderr

    Outputs:
        logger
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    # Set the handler
    if not logger.hasHandlers():
        handler = logging.StreamHandler(stream)
        formatter = logging.Formatter(
            "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger


###################################################################################
# System constants
#

r"""
This is a "generic" logger available to all scripts.
It is encouraged that each class has it's own custom logger with the name of the class.
You can use the "get_custom_logger" function to build a custom logger with a standard format.
"""
LOGGER = get_custom_logger("docling-pm", logging.INFO)

# Supported dataset types
supported_datasets = ["TF_prepared"]  # TF prepared dataset

# Split names
TRAIN_SPLIT = "train"
VAL_SPLIT = "val"
TEST_SPLIT = "test"

# Prepared data parts and filename templates
PREPARED_DATA_PARTS = {
    # Array with the bboxes (x1y1x2y2) for all cells of the images across all splits.
    # The bboxes are indexed with the filename.
    # Notices:
    #   - The bboxes are NOT transformed.
    #   - If the image filenames are the same across splits, there will be one one entry in the file
    "BBOXES": "BBOXES.json",
    # Image filenames used for train and val
    "IMAGES": "IMAGES.json",
    # Mean, std, variance as arrays of 3 (for each color)
    "STATISTICS": "STATISTICS_<POSTFIX>.json",  # PRECOMPUTED
    # Bboxes of the cells in the form [1, x1, x2, y1, y2] or [0, 0, 0, 0, 0] in case of no box.
    "TRAIN_CELLBBOXES": "TRAIN_CELLBBOXES_<POSTFIX>.json",  # NOT USED.
    # Array with arrays of the length + 2 of the original cells per image.
    "TRAIN_CELLLENS": "TRAIN_CELLLENS_<POSTFIX>.json",
    # Indices of the cells between <start> <end> and <pad> at the end.
    "TRAIN_CELLS": "TRAIN_CELLS_<POSTFIX>.json",
    # Array with the length + 2 of the original tags per image.
    "TRAIN_TAGLENS": "TRAIN_TAGLENS_<POSTFIX>.json",
    # Indices of the tags between <start> <end> and <pad> at the end.
    "TRAIN_TAGS": "TRAIN_TAGS_<POSTFIX>.json",
    # Ground truth for the evaluation dataset per eval image.
    "VAL": "VAL.json",
    # Vocabulary: Indices of the word_map_cells and word_map_tags
    "WORDMAP": "WORDMAP_<POSTFIX>.json",  # PRECOMPUTED
}

# Purposes
TRAIN_PURPOSE = "train"
VAL_PURPOSE = "val"
TEST_PURPOSE = "test"
PREDICT_PURPOSE = "predict"

# The DDP world size when we train in CPU with DDP enabled
DDP_CPU_WORLD_SIZE = 2



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import copy
import logging
import re

import numpy as np

import docling_ibm_models.tableformer.otsl as otsl
import docling_ibm_models.tableformer.settings as s

# LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG
LOG_LEVEL = logging.WARN

# Cell labels
BODY = "body"
COL_HEADER = "col_header"
MULTI_COL_HEADER = "multi_col_header"
MULTI_ROW_HEADER = "multi_row_header"
MULTI_ROW = "multi_row"
MULTI_COL = "multi_col"


def validate_bboxes_page(bboxes):
    r"""
    Useful function for Debugging

    Validate that the bboxes have a positive area in the page coordinate system

    Parameters
    ----------
    bboxes : list of 4
        Each element of the list is expected to be a bbox in the page coordinates system

    Returns
    -------
    int
        The number of invalid bboxes.
    """
    invalid_counter = 0
    for i, bbox in enumerate(bboxes):
        area = (bbox[2] - bbox[0]) * (bbox[3] - bbox[1])

        if area < 0:
            print("Wrong bbox: {} - {}".format(i, bbox))
            invalid_counter += 1

    if invalid_counter > 0:
        print("Invalid bboxes in total: {}".format(invalid_counter))
    return invalid_counter


def find_intersection(b1, b2):
    r"""
    Compute the intersection between 2 bboxes

    Parameters
    ----------
    b1 : list of 4
        The page x1y1x2y2 coordinates of the bbox
    b2 : list of 4
        The page x1y1x2y2 coordinates of the bbox

    Returns
    -------
    The bbox of the intersection or None if there is no intersection
    """
    # Check when the bboxes do NOT intersect
    if b1[2] < b2[0] or b2[2] < b1[0] or b1[1] > b2[3] or b2[1] > b2[3]:
        return None

    i_bbox = [
        max(b1[0], b2[0]),
        max(b1[1], b2[1]),
        min(b1[2], b2[2]),
        min(b1[3], b2[3]),
    ]
    return i_bbox


class CellMatcher:
    r"""
    Match the table cells to the pdf page cells.

    NOTICE: PDF page coordinate system vs table coordinate system.
    In both systems the bboxes are described in as (x1, y1, x2, y2) with the following meaning:

    Page coordinate system:
    - Origin (0, 0) at the lower-left corner
    - (x1, y1) the lower left corner of the box
    - (x2, y2) the upper right corner of the box

    Table coordinate system:
    - Origin (0, 0) at the upper-left corner
    - (x1, y1) the upper left corner of the box
    - (x2, y2) the lower right corner of the box
    """

    def __init__(self, config):
        self._config = config
        self._iou_thres = config["predict"]["pdf_cell_iou_thres"]

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def match_cells(self, iocr_page, table_bbox, prediction):
        r"""
        Convert the tablemodel prediction into the Docling format

        Parameters
        ----------
        iocr_page : dict
            The original Docling provided table data
        prediction : dict
            The dictionary has the keys:
            "tag_seq": The sequence in indices from the WORDMAP
            "html_seq": The sequence as html tags
            "bboxes": The bounding boxes

        Returns
        -------
        matching_details : dict
            Dictionary with all details about the mathings between the table and pdf cells
        """
        pdf_cells = copy.deepcopy(iocr_page["tokens"])
        if len(pdf_cells) > 0:
            for word in pdf_cells:
                if isinstance(word["bbox"], list):
                    continue
                elif isinstance(word["bbox"], dict):
                    word["bbox"] = [
                        word["bbox"]["l"],
                        word["bbox"]["t"],
                        word["bbox"]["r"],
                        word["bbox"]["b"],
                    ]
        table_bboxes = prediction["bboxes"]
        table_classes = prediction["classes"]
        # BBOXES transformed...
        table_bboxes_page = self._translate_bboxes(table_bbox, table_bboxes)

        # Combine the table tags and bboxes into TableCells
        html_seq = prediction["html_seq"]
        otsl_seq = prediction["rs_seq"]
        table_cells = self._build_table_cells(
            html_seq, otsl_seq, table_bboxes_page, table_classes
        )

        matches = {}
        matches_counter = 0
        if len(pdf_cells) > 0:
            matches, matches_counter = self._intersection_over_pdf_match(
                table_cells, pdf_cells
            )

        self._log().debug("matches_counter: {}".format(matches_counter))

        # Build output
        matching_details = {
            "iou_threshold": self._iou_thres,
            "table_bbox": table_bbox,
            "prediction_bboxes_page": table_bboxes_page,  # Make easier the comparison with c++
            "prediction": prediction,
            "pdf_cells": pdf_cells,
            "page_height": iocr_page["height"],
            "page_width": iocr_page["width"],
            "table_cells": table_cells,
            "pdf_cells": pdf_cells,
            "matches": matches,
        }
        return matching_details

    def match_cells_dummy(self, iocr_page, table_bbox, prediction):
        r"""
        Convert the tablemodel prediction into the Docling format
        DUMMY version doesn't do matching with text cells, but propagates predicted bboxes,
        respecting the rest of the format

        Parameters
        ----------
        iocr_page : dict
            The original Docling provided table data
        prediction : dict
            The dictionary has the keys:
            "tag_seq": The sequence in indices from the WORDMAP
            "html_seq": The sequence as html tags
            "bboxes": The bounding boxes

        Returns
        -------
        matching_details : dict
            Dictionary with all details about the mathings between the table and pdf cells
        """
        pdf_cells = copy.deepcopy(iocr_page["tokens"])
        if len(pdf_cells) > 0:
            for word in pdf_cells:
                word["bbox"] = [
                    word["bbox"]["l"],
                    word["bbox"]["t"],
                    word["bbox"]["r"],
                    word["bbox"]["b"],
                ]

        table_bboxes = prediction["bboxes"]
        table_classes = prediction["classes"]
        # BBOXES transformed...
        table_bboxes_page = self._translate_bboxes(table_bbox, table_bboxes)

        # Combine the table tags and bboxes into TableCells
        html_seq = prediction["html_seq"]
        otsl_seq = prediction["rs_seq"]

        table_cells = self._build_table_cells(
            html_seq, otsl_seq, table_bboxes_page, table_classes
        )

        # Build output
        matching_details = {
            "iou_threshold": self._iou_thres,
            "table_bbox": table_bbox,
            "prediction_bboxes_page": table_bboxes_page,
            "prediction": prediction,
            "pdf_cells": pdf_cells,
            "page_height": iocr_page["height"],
            "page_width": iocr_page["width"],
            "table_cells": table_cells,
            "pdf_cells": pdf_cells,
            "matches": {},
        }
        return matching_details

    def _build_table_cells(self, html_seq, otsl_seq, bboxes, table_classes):
        r"""
        Combine the tags and bboxes of the table into unified TableCell objects.
        Each TableCell takes a row_id, column_id index based on the html structure provided by
        html_seq.
        It is assumed that the bboxes are in sync with the appearence of the closing </td>

        Parameters
        ----------
        html_seq : list
            List of html tags
        bboxes : list of lists of 4
            Bboxes for the table cells at the page origin

        Returns
        -------
        list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        """
        table_html_structure = {
            "html": {"structure": {"tokens": html_seq}},
            "split": "predict",
            "filename": "memory",
        }

        otsl_spans = {}

        # r, o = otsl.html_to_otsl(table, writer, true, extra_debug, include_html)
        r, o = otsl.html_to_otsl(table_html_structure, None, False, False, True, False)
        if not r:
            ermsg = "ERR#: COULD NOT CONVERT TO RS THIS TABLE TO COMPUTE SPANS"
            self._log().debug(ermsg)
        else:
            otsl_spans = o["otsl_spans"]

        table_cells = []

        # It is assumed that the bboxes appear in sync (at the same order) as the TDs
        cell_id = 0

        row_id = -1
        column_id = -1
        in_header = False
        in_body = False
        multicol_tag = ""
        colspan_val = 0
        rowspan_val = 0

        mode = "OTSL"
        if mode == "HTML":
            for tag in html_seq:
                label = None
                if tag == "<thead>":
                    in_header = True
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                elif tag == "</thead>":
                    in_header = False
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                elif tag == "<tbody>":
                    in_body = True
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                elif tag == "</tbody>":
                    in_body = False
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                elif tag == "<td>" or tag == "<td":
                    column_id += 1
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                    if tag == "<td":
                        multicol_tag = tag
                elif tag == "<tr>":
                    row_id += 1
                    column_id = -1
                    multicol_tag = ""
                    colspan_val = 0
                    rowspan_val = 0
                elif "colspan" in tag:
                    label = MULTI_COL
                    multicol_tag += tag
                    colspan_val = int(re.findall(r'"([^"]*)"', tag)[0])
                elif "rowspan" in tag:
                    label = MULTI_ROW
                    multicol_tag += tag
                    rowspan_val = int(re.findall(r'"([^"]*)"', tag)[0])
                elif tag == "</td>":  # Create a TableCell on each closing td
                    if len(multicol_tag) > 0:
                        multicol_tag += tag
                    if in_header:
                        if label is None:
                            label = COL_HEADER
                        elif label == MULTI_COL:
                            label = MULTI_COL_HEADER
                        elif label == MULTI_ROW:
                            label = MULTI_ROW_HEADER
                    if label is None and in_body:
                        label = BODY

                    err_mismatch = "Mismatching bboxes with closing TDs {} < {}".format(
                        cell_id, len(bboxes)
                    )
                    assert cell_id < len(bboxes), err_mismatch
                    bbox = bboxes[cell_id]
                    cell_class = table_classes[cell_id]

                    table_cell = {}
                    table_cell["cell_id"] = cell_id
                    table_cell["row_id"] = row_id
                    table_cell["column_id"] = column_id
                    table_cell["bbox"] = bbox
                    table_cell["cell_class"] = cell_class
                    table_cell["label"] = label
                    table_cell["multicol_tag"] = multicol_tag
                    if colspan_val > 0:
                        table_cell["colspan_val"] = colspan_val
                        column_id += (
                            colspan_val - 1
                        )  # Shift column index to account for span
                    if rowspan_val > 0:
                        table_cell["rowspan_val"] = rowspan_val

                    table_cells.append(table_cell)
                    cell_id += 1

        if mode == "OTSL":
            row_id = 0
            column_id = 0
            multicol_tag = ""
            otsl_line = []
            cell_id_line = []

            for tag in otsl_seq:
                otsl_line.append(tag)
                if tag == "nl":
                    row_id += 1
                    column_id = 0
                    otsl_line = []
                    cell_id_line = []
                if tag in ["fcel", "ecel", "xcel", "ched", "rhed", "srow"]:
                    cell_id_line.append(cell_id)
                    bbox = [0.0, 0.0, 0.0, 0.0]
                    if cell_id < len(bboxes):
                        bbox = bboxes[cell_id]

                    cell_class = 2
                    if cell_id < len(table_classes):
                        cell_class = table_classes[cell_id]
                    label = tag

                    table_cell = {}
                    table_cell["cell_id"] = cell_id
                    table_cell["row_id"] = row_id
                    table_cell["column_id"] = column_id
                    table_cell["bbox"] = bbox
                    table_cell["cell_class"] = cell_class
                    table_cell["label"] = label
                    table_cell["multicol_tag"] = multicol_tag

                    colspan_val = 0
                    rowspan_val = 0

                    if cell_id in otsl_spans:
                        colspan_val = otsl_spans[cell_id][0]
                        rowspan_val = otsl_spans[cell_id][1]
                    if colspan_val > 0:
                        table_cell["colspan_val"] = colspan_val
                    if rowspan_val > 0:
                        table_cell["rowspan_val"] = rowspan_val

                    table_cells.append(table_cell)
                    cell_id += 1
                if tag != "nl":
                    column_id += 1

        return table_cells

    def _translate_bboxes(self, table_bbox, cell_bboxes):
        r"""
        Translate table cell bboxes to the lower-left corner of the page.

        The cells of the table are given:
        - Origin at the top left corner
        - Point A: Top left corner
        - Point B: Low right corner
        - Coordinate values are normalized to the table width/height

        Parameters
        ----------
        table_bbox : list of 4
            The whole table bbox page coordinates
        cell_bboxes : list of lists of 4
            The bboxes of the table cells

        Returns
        -------
        list of 4
            The translated bboxes of the table cells
        """
        W = table_bbox[2] - table_bbox[0]
        H = table_bbox[3] - table_bbox[1]
        b = np.asarray(cell_bboxes)
        t_mask = np.asarray(
            [table_bbox[0], table_bbox[3], table_bbox[0], table_bbox[3]]
        )
        m = np.asarray([W, -H, W, -H])
        page_bboxes_y_flipped = t_mask + m * b
        page_bboxes = page_bboxes_y_flipped[:, [0, 3, 2, 1]]  # Flip y1' with y2'
        page_bboxes_list = page_bboxes.tolist()

        t_height = table_bbox[3]
        page_bboxes_list1 = []
        for page_bbox in page_bboxes_list:
            page_bbox1 = [
                page_bbox[0],
                t_height - page_bbox[3] + table_bbox[1],
                page_bbox[2],
                t_height - page_bbox[1] + table_bbox[1],
            ]
            page_bboxes_list1.append(page_bbox1)
        return page_bboxes_list1

    def _intersection_over_pdf_match(self, table_cells, pdf_cells):
        r"""
        Compute Intersection between table cells and pdf cells,
        match 1 pdf cell with highest intersection with only 1 table cell.

        First compute and cache the areas for all involved bboxes.
        Then compute the pairwise intersections

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"

        pdf_cells : list of dict
            Each element of the list is a dictionary which should have the keys: "id", "bbox"
        Returns
        -------
        dictionary of lists of table_cells
            Return a dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell
        int
            Number of total matches
        """
        pdf_bboxes = np.asarray([p["bbox"] for p in pdf_cells])
        pdf_bboxes_areas = (pdf_bboxes[:, 2] - pdf_bboxes[:, 0]) * (
            pdf_bboxes[:, 3] - pdf_bboxes[:, 1]
        )

        # key: pdf_cell_id, value: list of TableCell that fall inside that pdf_cell
        matches = {}
        matches_counter = 0

        # Compute Intersections and build matches
        for i, table_cell in enumerate(table_cells):
            table_cell_id = table_cell["cell_id"]
            t_bbox = table_cell["bbox"]

            for j, pdf_cell in enumerate(pdf_cells):
                pdf_cell_id = pdf_cell["id"]
                p_bbox = pdf_cell["bbox"]

                # Compute intersection
                i_bbox = find_intersection(t_bbox, p_bbox)
                if i_bbox is None:
                    continue

                # Compute IOU and filter on threshold
                i_bbox_area = (i_bbox[2] - i_bbox[0]) * (i_bbox[3] - i_bbox[1])
                iopdf = 0
                if float(pdf_bboxes_areas[j]) > 0:
                    iopdf = i_bbox_area / float(pdf_bboxes_areas[j])

                if iopdf > 0:
                    match = {"table_cell_id": table_cell_id, "iopdf": iopdf}
                    if pdf_cell_id not in matches:
                        matches[pdf_cell_id] = [match]
                        matches_counter += 1
                    else:
                        # Check if the same match was not already counted
                        if match not in matches[pdf_cell_id]:
                            matches[pdf_cell_id].append(match)
                            matches_counter += 1
        return matches, matches_counter

    def _iou_match(self, table_cells, pdf_cells):
        r"""
        Use Intersection over Union to decide the matching between table cells and pdf cells

        First compute and cache the areas for all involved bboxes.
        Then compute the pairwise intersections and IOUs and keep those pairs that exceed the IOU
        threshold

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"

        pdf_cells : list of dict
            Each element of the list is a dictionary which should have the keys: "id", "bbox"
        Returns
        -------
        dictionary of lists of table_cells
            Return a dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell
        int
            Number of total matches
        """
        table_bboxes = np.asarray([t["bbox"] for t in table_cells])
        pdf_bboxes = np.asarray([p["bbox"] for p in pdf_cells])

        # Cache the areas for table bboxes and pdf bboxes
        table_bboxes_areas = (table_bboxes[:, 2] - table_bboxes[:, 0]) * (
            table_bboxes[:, 3] - table_bboxes[:, 1]
        )

        pdf_bboxes_areas = (pdf_bboxes[:, 2] - pdf_bboxes[:, 0]) * (
            pdf_bboxes[:, 3] - pdf_bboxes[:, 1]
        )

        # key: pdf_cell_id, value: list of TableCell that fall inside that pdf_cell
        matches = {}
        matches_counter = 0

        # Compute IOUs and build matches
        for i, table_cell in enumerate(table_cells):
            table_cell_id = table_cell["cell_id"]
            t_bbox = table_cell["bbox"]

            for j, pdf_cell in enumerate(pdf_cells):
                pdf_cell_id = pdf_cell["id"]
                pdf_cell_text = pdf_cell["text"]
                p_bbox = pdf_cell["bbox"]

                # Compute intersection
                i_bbox = find_intersection(t_bbox, p_bbox)
                if i_bbox is None:
                    continue

                # Compute IOU and filter on threshold
                i_bbox_area = (i_bbox[2] - i_bbox[0]) * (i_bbox[3] - i_bbox[1])
                iou = 0
                div_area = float(
                    table_bboxes_areas[i] + pdf_bboxes_areas[j] - i_bbox_area
                )
                if div_area > 0:
                    iou = i_bbox_area / div_area
                if iou < self._iou_thres:
                    continue

                if pdf_cell_id not in matches:
                    matches[pdf_cell_id] = []

                match = {
                    "table_cell_id": table_cell_id,
                    "iou": iou,
                    "text": pdf_cell_text,
                }
                matches[pdf_cell_id].append(match)
                matches_counter += 1

        return matches, matches_counter



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import numbers
from collections.abc import Iterable, Sequence

import cv2
import numpy as np
import torch
from torchvision.transforms import functional

cv2.setNumThreads(0)
cv2.ocl.setUseOpenCL(False)

INTER_MODE = {
    "NEAREST": cv2.INTER_NEAREST,
    "BILINEAR": cv2.INTER_LINEAR,
    "BICUBIC": cv2.INTER_CUBIC,
}

PAD_MOD = {
    "constant": cv2.BORDER_CONSTANT,
    "edge": cv2.BORDER_REPLICATE,
    "reflect": cv2.BORDER_DEFAULT,
    "symmetric": cv2.BORDER_REFLECT,
}


def _is_tensor_image(img):
    return torch.is_tensor(img) and img.ndimension() == 3


def _is_numpy_image(img):
    return isinstance(img, np.ndarray) and (img.ndim in {2, 3})


def normalize(tensor, mean, std):
    """Normalize a tensor image with mean and standard deviation.
    See ``Normalize`` for more details.
    Args:
        tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channely.
    Returns:
        Tensor: Normalized Tensor image.
    """
    if _is_tensor_image(tensor):
        for t, m, s in zip(tensor, mean, std, strict=False):
            t.sub_(m).div_(s)
        return tensor
    elif _is_numpy_image(tensor):
        return (tensor.astype(np.float32) - 255.0 * np.array(mean)) / np.array(std)
    else:
        raise RuntimeError("Undefined type")


def resize(img, size, interpolation="BILINEAR"):
    """Resize the input CV Image to the given size.
    Args:
        img (np.ndarray): Image to be resized.
        size (tuple or int): Desired output size. If size is a sequence like
            (h, w), the output size will be matched to this. If size is an int,
            the smaller edge of the image will be matched to this number maintaing
            the aspect ratio. i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (str, optional): Desired interpolation. Default is ``BILINEAR``
    Returns:
        cv Image: Resized image.
    """
    if not _is_numpy_image(img):
        raise TypeError("img should be CV Image. Got {}".format(type(img)))
    if not (isinstance(size, int) or (isinstance(size, Iterable) and len(size) == 2)):
        raise TypeError("Got inappropriate size arg: {}".format(size))

    # TODO(Nikos): Try to remove the opencv dependency
    if isinstance(size, int):
        h, w, c = img.shape
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return cv2.resize(
                img, dsize=(ow, oh), interpolation=INTER_MODE[interpolation]
            )
        else:
            oh = size
            ow = int(size * w / h)
            return cv2.resize(
                img, dsize=(ow, oh), interpolation=INTER_MODE[interpolation]
            )
    else:
        oh, ow = size
        return cv2.resize(
            img, dsize=(int(ow), int(oh)), interpolation=INTER_MODE[interpolation]
        )



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import time
from collections import deque
from statistics import mean, median

from docling_ibm_models.tableformer.utils.mem_monitor import MemMonitor


class SingletonClass(type):
    r"""
    Generic singleton metaclass
    """

    def __init__(self, name, bases, dic):
        self._instance = None
        super().__init__(name, bases, dic)

    def __call__(cls, *args, **kwargs):
        # Create a singleton if needed
        if cls._instance is None:
            singleton = cls.__new__(cls)
            singleton.__init__(*args, **kwargs)
            cls._instance = singleton
        return cls._instance


class Profiler:
    r"""
    Application specific profiler
    Decompose the application into "sections". Each section is a label.
    The total time a section consumes is split into "intervals"
    Use the `begin`, `end` methods to mark the begining and end of an interval for
    a certain section
    """

    def __init__(self):
        self._section_dts = {}  # section name -> sum(section intervals)
        self._section_calls = {}  # section name -> number of invocations
        self._section_kB = {}  # section name -> max kB of used heap (resident set size)

        # section name -> beginning of the last interval
        self._last_begin = {}

        self._mem_monitor = MemMonitor()

    def begin(self, section_name, enable=True):
        r"""
        Mark the beginning of an interval

        Parameters
        ----------
        section_name : string
            Name of the section
        enable : bool
            The actual interval entry takes place only if enable is true

        Return
        ------
            True if the interval has actuall begun
        """
        if not enable:
            return False
        self._last_begin[section_name] = time.time()
        return True

    def end(self, section_name, enable=True):
        r"""
        Mark the end of an interval for a certain section

        Parameters
        ----------
        section_name : string
            Name of the section
        enable : bool
            The actual interval entry takes place only if enable is true

        Return
        ------
            True if the section name is valid and an interval for this section has already begun
            False otherwise
        """
        if not enable:
            return False
        if section_name not in self._last_begin:
            return False

        # Get memory
        kB = self._mem_monitor.get_memory()
        if isinstance(kB, dict):
            kB = kB["resident"]

        dt = time.time() - self._last_begin[section_name]
        if section_name not in self._section_dts:
            self._section_dts[section_name] = dt
            self._section_calls[section_name] = 1
            self._section_kB[section_name] = kB
        else:
            self._section_dts[section_name] += dt
            self._section_calls[section_name] += 1
            self._section_kB[section_name] = max(kB, self._section_kB[section_name])

        return True

    def get_data(self, section_names=None):
        r"""
        Return a dict with profiling data for the specified sections.

        Parameter
        ---------
        section_names : list of string
            List with the section names to get their accumulative dt
            If it is None, all sections are returned

        Return
        ------
        dict of dicts
            Outer key: section name
            Inner keys: "dt": Accumulative time for that section, "cells": Number of calls
        """
        # Filter the section names to apply
        filtered_names = list(
            filter(lambda x: x in section_names, self._section_dts.keys())
            if section_names is not None
            else self._section_dts.keys()
        )
        data = {}
        for section_name in filtered_names:
            data[section_name] = {
                "dt": self._section_dts[section_name],
                "calls": self._section_calls[section_name],
                "kB": self._section_kB[section_name],
            }
        return data


class AppProfiler(Profiler, metaclass=SingletonClass):
    r"""
    AppProfiler is a singleton of the Profiler for application wide usage
    """

    def __init__(self):
        super(AppProfiler, self).__init__()


class AggProfiler(metaclass=SingletonClass):
    r"""
    Generic wrapper of Profiler that enables aggregation of profiling statistics around Cycles

    - When a new cycle begins a new Profiler is created to keep the profiling data per section
    - Keep the last n cycles in a sliding window manner
    - At every time we can get profiling data about the last cycle and statistics over the last n
      cycles
    """

    def __init__(self, window_size=20):
        self._window_size = window_size
        # deque with up to the last "window_size" Profilers. The newest at index 0
        self._cycles = deque()

    def start_agg(self, enable=True):
        r"""
        Returns
        -------
        0: not enabled
        1: a new scope has started
        """
        if not enable:
            return 0

        # Add a new profiler
        self._cycles.appendleft(Profiler())
        # In case the deque has grown too much, remove the oldest Profiler
        if len(self._cycles) > self._window_size:
            self._cycles.pop()
        return 1

    def begin(self, section_name, enable=True):
        if not enable:
            return False
        if len(self._cycles) == 0:
            print("AggProfiler begin | Start Aggregator not initialized.")
            return False
        profiler = self._cycles[0]
        return profiler.begin(section_name)

    def end(self, section_name, enable=True):
        if not enable:
            return False
        if len(self._cycles) == 0:
            print("AggProfiler end | Start Aggregator not initialized.")
            return False
        profiler = self._cycles[0]
        return profiler.end(section_name)

    def get_data(self):
        r"""
        Get profiling data for:
        - The last cycle
        - Aggragated statistics (avg, median) per section and per metric across all cycles
        - The dt numbers for the mean/median is the average time for each section ACROSS the cycle
        - There is NO need to compute average by yourself.

        Returns
        -------
        dict with the structure:
        - window: int with the size of the time sliding window
        - last: dict with the metrics for the last cycle (as provided by the Profiler)
        - mean: dict with the mean metrics per section across the cycle
            - section_name
                - metric_name: mean of the metric values
        - median: dict with the median metrics per section across the cycle
            - section_name
                - metric_name: median of the metric values
        """
        last_data = self._cycles[0].get_data()
        data = {
            "window": len(self._cycles),
            "last": last_data,
            "mean": {},
            "median": {},
        }

        # Section -> metric -> [values]
        section_metric_values = {}

        # Collect the metrics
        for i, p in enumerate(self._cycles):
            p_data = p.get_data()
            for section_name, m_dict in p_data.items():
                for m_name, m_val in m_dict.items():
                    if section_name not in section_metric_values:
                        section_metric_values[section_name] = {}
                    s_metrics = section_metric_values[section_name]
                    if m_name not in s_metrics:
                        s_metrics[m_name] = []
                    s_metrics[m_name].append(m_val)

        # Aggregate the metrics
        for section_name, m_dict in section_metric_values.items():
            for m_name, m_values in m_dict.items():
                if section_name not in data["mean"]:
                    data["mean"][section_name] = {}
                if section_name not in data["median"]:
                    data["median"][section_name] = {}

                mean_v = mean(m_values)
                median_v = median(m_values)
                data["mean"][section_name][m_name] = mean_v
                data["median"][section_name][m_name] = median_v

        return data



# Copyright 2023 Haotian Liu
#
# This file is part of the Vary project, originally located at:
# https://github.com/Ucas-HaoranWei/Vary-toy/blob/main/Vary-master/vary/model/vary_opt.py
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
from transformers import (
    AutoConfig,
    AutoModelForCausalLM,
    OPTConfig,
    OPTForCausalLM,
    OPTModel,
)
from transformers.modeling_outputs import (
    BaseModelOutputWithPast,
    CausalLMOutputWithPast,
)

from docling_ibm_models.code_formula_model.models.sam import build_sam_vit_b


class SamOptConfig(OPTConfig):
    model_type = "sam_opt"

    def __init__(
        self,
        sam_image_size=1024,
        sam_mm_projector_in=1024,
        sam_mm_projector_out=768,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sam_image_size = sam_image_size
        self.sam_mm_projector_in = sam_mm_projector_in
        self.sam_mm_projector_out = sam_mm_projector_out


class SamOPTModel(OPTModel):
    config_class = SamOptConfig

    def __init__(self, config: OPTConfig):
        super(SamOPTModel, self).__init__(config)
        self.vision_tower = build_sam_vit_b(image_size=config.sam_image_size)

        self.mm_projector = nn.Linear(
            config.sam_mm_projector_in, config.sam_mm_projector_out
        )

    def embed_tokens(self, x):
        return self.get_input_embeddings()(x)

    def forward(
        self,
        input_ids: torch.LongTensor,
        attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[torch.FloatTensor] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, BaseModelOutputWithPast]:

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        vision_tower = getattr(self, "vision_tower", None)
        im_start_token = getattr(self.config, "im_start_token", -1)

        if input_ids.shape[1] != 1 or self.training:
            with torch.set_grad_enabled(self.training):
                assert vision_tower is not None
                image_features = vision_tower(images)
                image_features = image_features.flatten(2).permute(0, 2, 1)
                image_features = self.mm_projector(image_features)

            new_input_embeds = []
            for cur_input_ids, cur_input_embeds, cur_image_features in zip(
                input_ids, inputs_embeds, image_features
            ):
                image_start_token_position = int(
                    torch.where(cur_input_ids == im_start_token)[0].item()
                )  # cast to int for mypy

                cur_image_features = cur_image_features.to(
                    device=cur_input_embeds.device
                )
                num_patches = cur_image_features.shape[0]
                cur_input_embeds = torch.cat(
                    (
                        cur_input_embeds[: image_start_token_position + 1],
                        cur_image_features,
                        cur_input_embeds[
                            image_start_token_position + num_patches + 1 :
                        ],
                    ),
                    dim=0,
                )

                new_input_embeds.append(cur_input_embeds)

            inputs_embeds = torch.stack(new_input_embeds, dim=0)  # type: ignore

        return super(SamOPTModel, self).forward(
            input_ids=None,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )


class SamOPTForCausalLM(OPTForCausalLM):
    config_class = SamOptConfig

    def __init__(self, config):
        super(OPTForCausalLM, self).__init__(config)
        self.model = SamOPTModel(config)

        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        self.post_init()

    def get_model(self):
        return self.model

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[torch.FloatTensor] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )

        outputs = self.model(
            input_ids=input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            images=images,
            return_dict=return_dict,
        )

        hidden_states = outputs[0]
        logits = self.lm_head(hidden_states).contiguous()

        return CausalLMOutputWithPast(
            loss=None,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def prepare_inputs_for_generation(
        self, input_ids, past_key_values=None, inputs_embeds=None, **kwargs
    ):
        token_type_ids = kwargs.get("token_type_ids", None)
        if past_key_values:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past_key_values:
                position_ids = position_ids[:, -1].unsqueeze(-1)
        else:
            position_ids = None

        if inputs_embeds is not None and past_key_values is None:
            model_inputs = {"inputs_embeds": inputs_embeds}
        else:
            model_inputs = {"input_ids": input_ids}

        model_inputs.update(
            {
                "past_key_values": past_key_values,
                "use_cache": kwargs.get("use_cache"),
                "position_ids": position_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,
                "images": kwargs.get("images", None),
            }
        )
        return model_inputs


AutoConfig.register("sam_opt", SamOptConfig)
AutoModelForCausalLM.register(SamOptConfig, SamOPTForCausalLM)



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
from PIL import Image
from torchvision.transforms import functional as F
from transformers import AutoImageProcessor
from transformers.image_processing_utils import ImageProcessingMixin


class SamOptImageProcessor(ImageProcessingMixin):

    def __init__(self, size=(1024, 1024), mean=None, std=None, **kwargs):
        super().__init__(**kwargs)
        self.size = size
        self.mean = mean
        self.std = std

    def __call__(self, image):
        if not isinstance(image, Image.Image):
            raise ValueError("Input must be a PIL Image")

        image = F.resize(image, self.size)
        image = F.to_tensor(image)

        image = F.normalize(image, mean=self.mean, std=self.std)

        return image


AutoImageProcessor.register(
    config_class="SamOptImageProcessor",
    slow_image_processor_class=SamOptImageProcessor,
)
